# eid.py - functions for handling Belgian Identity Card Number.
# coding: utf-8
#
# Copyright (C) 2024 Jeff Horemans
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""eID Number (Belgian electronic Identity Card Number).

The eID is an electronic identity card (with chip), issued to Belgian citizens
over 12 years old.
The card number applies only to the card in question and should not be confused
with the Belgian National Number (Rijksregisternummer, Numéro National),
that is also included on the card.

The card number consists of 12 digits in the form xxx-xxxxxxx-yy where
yy is a check digit calculated as the remainder of dividing xxxxxxxxxx by 97.
If the remainder is 0, the check number is set to 97.

More information:

* https://www.ibz.rrn.fgov.be/nl/identiteitsdocumenten/eid/
* https://eid.belgium.be/en/what-eid
* https://www.ibz.rrn.fgov.be/fileadmin/user_upload/nl/rr/instructies/IT-lijst/IT195_Identiteitsbewijs_20240115.pdf

>>> compact('000-0011032-71')
'000001103271'
>>> compact('591-1917064-58')
'591191706458'
>>> validate('000-0011032-71')
'000001103271'
>>> validate('591-1917064-58')
'591191706458'
>>> validate('591-2010999-97')
'591201099997'
>>> validate('000-0011032-25')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> format('591191706458')
'591-1917064-58'
"""

from __future__ import annotations

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number: str) -> str:
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -./').upper().strip()


def _calc_check_digits(number: str) -> str:
    """Calculate the expected check digits for the number, calculated as
    the remainder of dividing the first 10 digits of the number by 97.
    If the remainder is 0, the check number is set to 97.
    """
    return '%02d' % ((int(number[:10]) % 97) or 97)


def validate(number: str) -> str:
    """Check if the number is a valid ID card number.
    This checks the length, formatting and check digit."""
    number = compact(number)
    if not isdigits(number) or int(number) <= 0:
        raise InvalidFormat()
    if len(number) != 12:
        raise InvalidLength()
    if _calc_check_digits(number[:-2]) != number[-2:]:
        raise InvalidChecksum()
    return number


def is_valid(number: str) -> bool:
    """Check if the number is a valid Belgian ID Card number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number: str) -> str:
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return '-'.join((number[:3], number[3:10], number[10:]))



# nuit.py - functions for handling Mozambique NUIT numbers
# coding: utf-8
#
# Copyright (C) 2023 Leandro Regueiro
# Copyright (C) 2025 Luca Sicurello
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""NUIT (Número Único de Identificação Tributaria, Mozambique tax number).

This number consists of 9 digits, sometimes separated in three groups of three
digits using whitespace to make it easier to read.

The first digit indicates the type of entity. The next seven digits are a
sequential number. The last digit is the check digit, which is used to verify
the number was correctly typed.

More information:

* https://www.mobilize.org.mz/nuit-numero-unico-de-identificacao-tributaria/
* https://www.at.gov.mz/por/Perguntas-Frequentes2/NUIT

>>> validate('400339910')
'400339910'
>>> validate('400 005 834')
'400005834'
>>> validate('12345')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> format('400339910')
'400 339 910'
"""

from __future__ import annotations

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number: str) -> str:
    """Convert the number to the minimal representation."""
    return clean(number, ' -.').strip()


def calc_check_digit(number: str) -> str:
    """Calculate the check digit."""
    weights = (8, 9, 4, 5, 6, 7, 8, 9)
    check = sum(w * int(n) for w, n in zip(weights, number)) % 11
    return '01234567891'[check]


def validate(number: str) -> str:
    """Check if the number is a valid Mozambique NUIT number."""
    number = compact(number)
    if len(number) != 9:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    if calc_check_digit(number[:-1]) != number[-1]:
        raise InvalidChecksum()
    return number


def is_valid(number: str) -> bool:
    """Check if the number is a valid Mozambique NUIT number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number: str) -> str:
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return ' '.join([number[:3], number[3:-3], number[-3:]])



# npwp.py - functions for handling Indonesian NPWP numbers
# coding: utf-8
#
# Copyright (C) 2020 Leandro Regueiro
# Copyright (C) 2024 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""NPWP (Nomor Pokok Wajib Pajak, Indonesian VAT Number).

The Nomor Pokok Wajib Pajak (NPWP) is assigned to organisations and
individuals (families) by the Indonesian Tax Office after registration by the
tax payers.

The number consists of 16 digits and is either a NIK (Nomor Induk Kependudukan)
or a number that starts with a 0, followed by 2 digits that denote the type of
entity, 6 digits to identify the tax payer, a check digit over the first 8
digits followed by 3 digits to identify the local tax office and 3 digits for
branch code.

This module also accepts the old 15 digit format which is just the 16 digit
format without the starting 0.

More information:

* https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/Indonesia-TIN.pdf
* https://metacpan.org/pod/Business::ID::NPWP
* https://wiki.scn.sap.com/wiki/display/CRM/Indonesia

>>> validate('01.312.166.0-091.000')
'013121660091000'
>>> validate('016090524017000')
'016090524017000'
>>> validate('123456789')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> format('013000666091000')
'01.300.066.6-091.000'
"""  # noqa: E501

from __future__ import annotations

from stdnum import luhn
from stdnum.exceptions import *
from stdnum.id import nik
from stdnum.util import clean, isdigits


def compact(number: str) -> str:
    """Convert the number to the minimal representation.

    This strips the number of any valid separators and removes
    surrounding whitespace.
    """
    return clean(number, ' -.').strip()


def validate(number: str) -> str:
    """Check if the number is a valid Indonesia NPWP number."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) == 15:
        # Old 15 digit format
        luhn.validate(number[:9])
        return number
    if len(number) == 16:
        # New format since 2024: either a NIK (for Indonesian citizens) or
        # the old number with a 0 at the beginning
        if not number.startswith('0'):
            return nik.validate(number)
        luhn.validate(number[:10])
        return number
    raise InvalidLength()


def is_valid(number: str) -> bool:
    """Check if the number is a valid Indonesia NPWP number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number: str) -> str:
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return '%s.%s.%s.%s-%s.%s' % (
        number[:2], number[2:5], number[5:8], number[8], number[9:12], number[12:])



# ogm_vcs.py - functions for handling Belgian OGM-VCS
# coding: utf-8
#
# Copyright (C) 2025 Cédric Krier
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Belgian OGM-VCS.

The OGM-VCS is used in bank transfer as structured communication.

* https://febelfin.be/en/publications/2023/febelfin-banking-standards-for-online-banking

>>> validate('+++010/8068/17183+++')
'010806817183'
>>> validate('010/8068/17180')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> format('010806817183')
'010/8068/17183'
>>> calc_check_digits('0108068171')
'83'
"""

from __future__ import annotations

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number: str) -> str:
    """Convert the number to the minimal representation. This strips the number
    of any invalid separators and removes surrounding whitespace."""
    return clean(number, ' +/').strip()


def calc_check_digits(number: str) -> str:
    """Calculate the check digit that should be added."""
    number = compact(number)
    return '%02d' % ((int(number[:10]) % 97) or 97)


def validate(number: str) -> str:
    """Check if the number is a valid OGM-VCS."""
    number = compact(number)
    if not isdigits(number) or int(number) <= 0:
        raise InvalidFormat()
    if len(number) != 12:
        raise InvalidLength()
    if calc_check_digits(number) != number[-2:]:
        raise InvalidChecksum()
    return number


def is_valid(number: str) -> bool:
    """Check if the number is a valid VAT number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number: str) -> str:
    """Format the number provided for output."""
    number = compact(number)
    number = number.rjust(12, '0')
    return f'{number[:3]}/{number[3:7]}/{number[7:]}'



# cae.py - functions for handling Spanish CAE number
# coding: utf-8
#
# Copyright (C) 2024 Quique Porta
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""CAE (Código de Actividad y Establecimiento, Spanish activity establishment code).

The Código de Actividad y Establecimiento (CAE) is assigned by the Spanish
Tax Agency companies or establishments that carry out activities related to
products subject to excise duty. It identifies an activity and the
establishment in which it is carried out.

The number consists of 13 characters where the sixth and seventh characters
identify the managing office in which the territorial registration is carried
out and the eighth and ninth characters identify the activity that takes
place.


More information:

* https://www.boe.es/boe/dias/2006/12/28/pdfs/A46098-46100.pdf
* https://www2.agenciatributaria.gob.es/L/inwinvoc/es.aeat.dit.adu.adce.cae.cw.AccW?fAccion=consulta

>>> validate('ES00008V1488Q')
'ES00008V1488Q'
>>> validate('00008V1488')  # invalid check length
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> is_valid('ES00008V1488Q')
True
>>> is_valid('00008V1488')
False
>>> compact('ES00008V1488Q')
'ES00008V1488Q'
"""

from __future__ import annotations

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


_OFFICES = {
    '01',  # Álava
    '02',  # Albacete
    '03',  # Alicante
    '04',  # Almería
    '05',  # Ávila
    '06',  # Badajoz
    '07',  # Illes Balears
    '08',  # Barcelona
    '09',  # Burgos
    '10',  # Cáceres
    '11',  # Cádiz
    '12',  # Castellón
    '13',  # Ciudad Real
    '14',  # Córdoba
    '15',  # A Coruña
    '16',  # Cuenca
    '17',  # Girona
    '18',  # Granada
    '19',  # Guadalajara
    '20',  # Guipúzcoa
    '21',  # Huelva
    '22',  # Huesca
    '23',  # Jaén
    '24',  # León
    '25',  # Lleida
    '26',  # La Rioja
    '27',  # Lugo
    '28',  # Madrid
    '29',  # Málaga
    '30',  # Murcia
    '31',  # Navarra
    '32',  # Ourense
    '33',  # Oviedo
    '34',  # Palencia
    '35',  # Las Palmas
    '36',  # Pontevedra
    '37',  # Salamanca
    '38',  # Santa Cruz de Tenerife
    '39',  # Santander
    '40',  # Segovia
    '41',  # Sevilla
    '42',  # Soria
    '43',  # Tarragona
    '44',  # Teruel
    '45',  # Toledo
    '46',  # Valencia
    '47',  # Valladolid
    '48',  # Bizcaia
    '49',  # Zamora
    '50',  # Zaragoza
    '51',  # Cartagena
    '52',  # Gijón
    '53',  # Jerez de la Frontera
    '54',  # Vigo
    '55',  # Ceuta
    '56',  # Melilla
}

_ACTIVITY_KEYS = {
    'A1',  # Fábricas de alcohol
    'B1',  # Fábricas de bebidas derivadas
    'B9',  # Elaboradores de productos intermedios distintos de los comprendidos en B-0
    'B0',  # Elaboradores de productos intermedios en régimen especial
    'BA',  # Fábricas de bebidas alcohólicas
    'C1',  # Fábricas de cerveza
    'DA',  # Destiladores artesanales
    'EC',  # Fábricas de extractos y concentrados alcohólicos
    'F1',  # Elaboradores de otras bebidas fermentadas
    'V1',  # Elaboradores de vinos
    'A7',  # Depósitos fiscales de alcohol
    'AT',  # Almacenes fiscales de alcohol
    'B7',  # Depósitos fiscales de bebidas derivadas
    'BT',  # Almacenes fiscales de bebidas alcohólicas
    'C7',  # Depósitos fiscales de cerveza
    'DB',  # Depósitos fiscales de bebidas alcohólicas
    'E7',  # Depósitos fiscales de extractos y concentrados alcohólicos exclusivamente
    'M7',  # Depósitos fiscales de productos intermedios
    'OA',  # Operadores registrados de alcohol
    'OB',  # Operadores registrados de bebidas alcohólicas
    'OE',  # Operadores registrados de extractos y concentrados alcohólicos
    'OV',  # Operadores registrados de vinos y de otras bebidas fermentadas
    'V7',  # Depósitos fiscales de vinos y de otras bebidas fermentadas
    'B6',  # Plantas embotelladoras de bebidas derivadas
    'A2',  # Centros de investigación
    'A6',  # Usuarios de alcohol totalmente desnaturalizado
    'A9',  # Industrias de especialidades farmacéuticas
    'A0',  # Centros de atención médica
    'AC',  # Usuarios con derecho a devolución
    'AV',  # Usuarios de alcohol parcialmente desnaturalizado con desnaturalizante general
    'AW',  # Usuarios de alcohol parcialmente desnaturalizado con desnaturalizante especial
    'AX',  # Fábricas de vinagre
    'H1',  # Refinerías de crudo de petróleo
    'H2',  # Fábricas de biocarburante, consistente en alcohol etílico
    'H4',  # Fábricas de biocarburante o biocombustible con sistente en biodiesel
    'H6',  # Fábricas de biocarburante o biocombustible con sistente en alcohol metílico
    'H9',  # Industrias extractoras de gas natural y otros productos gaseosos
    'H0',  # Las demás industrias que obtienen productos gravados
    'HD',  # Industrias o establecimientos que someten productos a un tratamiento definido o,
           # Previa solicitud, a una transformación química
    'HH',  # Industrias extractoras de crudo de petróleo
    'H7',  # Depósitos fiscales de hidrocarburos
    'H8',  # Depósitos fiscales exclusivamente de biocarburantes
    'HB',  # Obtención accesoria de productos sujetos alimpuesto
    'HF',  # Almacenes fiscales para el suministro directo a instalaciones fijas
    'HI',  # Depósitos fiscales exclusivamente para la distribución de querosenos y gasolinas de aviación
    'HJ',  # Depósitos fiscales exclusivamente de productos de la tarifa segunda
    'HK',  # Instalaciones de venta de gas natural con tipo general y tipo reducido
    'HL',  # Almacenes fiscales exclusivamente de productos de la tarifa segunda
    'HM',  # Almacenes fiscales para la gestión de aceites usados destinados a su utilización como  combustibles
    'HN',  # Depósitos fiscales constituidos por una red de oleoductos
    'HT',  # Almacenes fiscales para el comercio al por mayor de hidrocarburos
    'HU',  # Almacenes fiscales constituidos por redes de transporte o distribución de gas natural
    'HV',  # Puntos de suministro marítimo de gasóleo
    'HX',  # Depósitos fiscales constituidos por una red de gasoductos
    'HZ',  # Detallistas de gasóleo
    'OH',  # Operadores registrados de hidrocarburos
    'HA',  # Titulares de aeronaves que utilizan instalaciones privadas
    'HC',  # Explotaciones industriales y proyectos piloto con derecho a devolución
    'HE',  # Los demás usuarios con derecho a exención
    'HP',  # Inyección en altos hornos
    'HQ',  # Construcción, modificación, pruebas y mantenimiento de aeronaves y embarcaciones
    'HR',  # Producción de electricidad en centrales eléctricas o producción de electricidad o
           # cogeneración de electricidad y de calor en centrales combinadas
    'HS',  # Transporte por ferrocarril
    'HW',  # Consumidores de combustibles y carburantes a tipo reducido (artículos 106.4 y 108
           # del Reglamento de los Impuestos Especiales)
    'T1',  # Fábricas de labores del tabaco
    'OT',  # Operadores registrados de labores del tabaco
    'T7',  # Depósitos fiscales de labores del tabaco
    'TT',  # Almacenes fiscales de labores del tabaco
    'L1',  # Fábricas de electricidad en régimen ordinario
    'L2',  # Generadores o conjunto de generadores de potencia total superior a 100 kilovatios
    'L0',  # Fábricas de electricidad en régimen especial
    'L3',  # Los demás sujetos pasivos
    'L7',  # Depósitos fiscales de electricidad
    'AF',  # Almacenes fiscales de bebidas alcohólicas y de labores del tabaco
    'DF',  # Depósitos fiscales de bebidas alcohólicas y de labores del tabaco
    'DM',  # Depósitos fiscales de bebidas alcohólicas y de labores del tabaco situados en
           # puertos y aeropuertos y que funcionen exclusivamente como establecimientos minoristas
    'DP',  # Depósitos fiscales para el suministro de bebidas alcohólicas y de labores del
           # tabaco para consumo o venta a bordo de buques y/o aeronaves
    'OR',  # Operadores registrados de bebidas alcohólicas y de labores del tabaco
    'PF',  # Industrias o usuarios en régimen de perfeccionamiento fiscal
    'RF',  # Representantes fiscales
    'VD',  # Empresas de ventas a distancia
}


def compact(number: str) -> str:
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number).upper().strip()


def validate(number: str) -> str:
    """Check if the number provided is a valid CAE number. This checks the
    length and formatting."""
    number = compact(number)
    if len(number) != 13:
        raise InvalidLength()
    if number[:2] != 'ES':
        raise InvalidFormat()
    if number[2:5] != '000':
        raise InvalidFormat()
    if number[5:7] not in _OFFICES:
        raise InvalidFormat()
    if number[7:9] not in _ACTIVITY_KEYS:
        raise InvalidFormat()
    if not isdigits(number[9:12]):
        raise InvalidFormat()
    if not number[12].isalpha():
        raise InvalidFormat()
    return number


def is_valid(number: str) -> bool:
    """Check if the number provided is a valid CAE number. This checks the
    length and formatting."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# accise.py - functions for handling French Accise numbers
# coding: utf-8
#
# Copyright (C) 2023 Cédric Krier
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""n° d'accise (French number to identify taxpayers of excise taxes).

The n° d'accise always start by FR0 following by the 2 ending digits of the
year, 3 number of customs office, one letter for the type and an ordering
number of 4 digits.

>>> validate('FR023004N9448')
'FR023004N9448'
>>> validate('FR0XX907E0820')
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> validate('FR012907A0820')
Traceback (most recent call last):
    ...
InvalidComponent: ...
"""

from __future__ import annotations

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


OPERATORS = set(['E', 'N', 'C', 'B'])


def compact(number: str) -> str:
    """Convert the number to the minimal representation. This strips the number
    of any valid separators and removes surrounding whitespace."""
    return clean(number, ' ').upper().strip()


def validate(number: str) -> str:
    """Check if the number is a valid accise number. This checks the length,
    formatting."""
    number = compact(number)
    code = number[:3]
    if code != 'FR0':
        raise InvalidFormat()
    if len(number) != 13:
        raise InvalidLength()
    if not isdigits(number[3:5]):
        raise InvalidFormat()
    if not isdigits(number[5:8]):
        raise InvalidFormat()
    if number[8] not in OPERATORS:
        raise InvalidComponent()
    if not isdigits(number[9:12]):
        raise InvalidFormat()
    return number


def is_valid(number: str) -> bool:
    """Check if the number is a valid accise number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# nace.py - functions for handling EU NACE classification
# coding: utf-8
#
# Copyright (C) 2017-2025 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""NACE (classification for businesses in the European Union).

The NACE (nomenclature statistique des activités économiques dans la
Communauté européenne) is a 4-level (and up to 4 digit) code for classifying
economic activities. It is the European implementation of the UN
classification ISIC.

The first 4 digits are the same in all EU countries while additional levels
and digits may be vary between countries. This module validates the numbers
according to revision 2.1 (or 2.0 if that version is supplied) and based on
the registry as published by the EC.

More information:

* https://en.wikipedia.org/wiki/Statistical_Classification_of_Economic_Activities_in_the_European_Community
* https://ec.europa.eu/eurostat/ramon/nomenclatures/index.cfm?TargetUrl=LST_NOM_DTL&StrNom=NACE_REV2&StrLanguageCode=EN&IntPcKey=&StrLayoutCode=HIERARCHIC
* https://showvoc.op.europa.eu/#/datasets/ESTAT_Statistical_Classification_of_Economic_Activities_in_the_European_Community_Rev._2.1._%28NACE_2.1%29/data

>>> validate('A')
'A'
>>> validate('62.01', revision='2.0')
'6201'
>>> get_label('62.10')
'Computer programming activities'
>>> validate('62.05')
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> validate('62059')  # does not validate country-specific numbers
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> format('6201')
'62.01'
"""  # noqa: E501

from __future__ import annotations

import warnings

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


# The revision of the NACE definition to use
DEFAULT_REVISION = '2.1'


def compact(number: str) -> str:
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, '.').strip()


def info(number: str, *, revision: str = DEFAULT_REVISION) -> dict[str, str]:
    """Lookup information about the specified NACE. This returns a dict."""
    number = compact(number)
    revision = revision.replace('.', '')
    from stdnum import numdb
    info = dict()
    for _n, i in numdb.get(f'eu/nace{revision}').info(number):
        if not i:
            raise InvalidComponent()
        info.update(i)
    return info


def get_label(number: str, revision: str = DEFAULT_REVISION) -> str:
    """Lookup the category label for the number."""
    return info(number, revision=revision)['label']


def label(number: str) -> str:  # pragma: no cover (deprecated function)
    """DEPRECATED: use `get_label()` instead."""  # noqa: D40
    warnings.warn(
        'label() has been to get_label()',
        DeprecationWarning, stacklevel=2)
    return get_label(number, revision='2.0')


def validate(number: str, revision: str = DEFAULT_REVISION) -> str:
    """Check if the number is a valid NACE. This checks the format and
    searches the registry to see if it exists."""
    number = compact(number)
    if len(number) > 4:
        raise InvalidLength()
    elif len(number) == 1:
        if not number.isalpha():
            raise InvalidFormat()
    else:
        if not isdigits(number):
            raise InvalidFormat()
    info(number, revision=revision)
    return number


def is_valid(number: str, revision: str = DEFAULT_REVISION) -> bool:
    """Check if the number is a valid NACE. This checks the format and
    searches the registry to see if it exists."""
    try:
        return bool(validate(number, revision=revision))
    except ValidationError:
        return False


def format(number: str) -> str:
    """Reformat the number to the standard presentation format."""
    return '.'.join((number[:2], number[2:])).strip('.')



#!/usr/bin/env python3

# update/cfi.py - script to download CFI code list from the SIX group
#
# Copyright (C) 2022-2026 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""This script downloads the list of CFI codes as published by the SIX group."""

import io
import re

import lxml.html
import openpyxl
import requests


# the location of the Statistical Classification file
download_url = 'https://www.six-group.com/en/products-services/financial-information/data-standards.html'


# The user agent that will be passed in requests
user_agent = 'Mozilla/5.0 (compatible; python-stdnum updater; +https://arthurdejong.org/python-stdnum/)'


def normalise(value):
    """Clean and minimise attribute names and values."""
    return re.sub(r' *[(\[\n].*', '', value, flags=re.MULTILINE).strip()


def get_categories(sheet):
    """Get the list of top-level CFI categories."""
    for row in sheet.iter_rows():
        if row[0].value and len(row[0].value) == 1 and row[1].value:
            yield (row[0].value, row[1].value)


def get_attributes(sheet):
    """Get the list of characters and attributes from the group-specific sheet."""
    attribute = None
    value_list = []
    values = None
    for row in sheet.iter_rows():
        if row[0].value and not row[1].value and row[2].value:
            attribute = normalise(row[2].value)
            values = []
            value_list.append((attribute, values))
        elif attribute and row[1].value and row[2].value:
            values.append((row[1].value, normalise(row[2].value)))
    return value_list


def print_attributes(attributes, index=0):
    """Print the collected attributes in a nested structure."""
    attribute, values = attributes[index]
    if len(values) == 1 and values[0][0] == 'X':
        print('%sA-Z' % (' ' * (index + 2)))
    else:
        for char, value in sorted(values):
            print('%s%s v="%s"' % (' ' * (index + 2), char, value))
        print('%sA-Z a="%s"' % (
            ' ' * (index + 2), attribute))
    if index < 3:
        print_attributes(attributes, index + 1)


if __name__ == '__main__':
    # Download the page that contains the link to the current XLS file
    response = requests.get(download_url, timeout=30, headers={'User-Agent': user_agent})
    response.raise_for_status()
    # Find the download link
    document = lxml.html.document_fromstring(response.content)
    links = [a.get('href') for a in document.findall('.//a[@href]')]
    link_url = next(a for a in links if re.match(r'.*/cfi/.*xlsx?$', a))
    # Download and parse the spreadsheet
    response = requests.get(link_url, timeout=30, headers={'User-Agent': user_agent})
    response.raise_for_status()
    workbook = openpyxl.load_workbook(io.BytesIO(response.content), read_only=True)

    print('# generated from %s, downloaded from' % link_url.split('/')[-1])
    print('# %s' % download_url)

    groups = sorted(x for x in workbook.sheetnames if len(x) == 6 and x.endswith('XXXX'))
    for category, name in sorted(get_categories(workbook['Categories'])):
        print('%s category="%s"' % (category, name))
        for group in (x for x in groups if x.startswith(category)):
            sheet = workbook[group]
            print(' %s group="%s"' % (group[1], normalise(sheet.cell(1, 1).value)))
            print_attributes(get_attributes(sheet))



#!/usr/bin/env python3
# coding: utf-8

# update/at_postleitzahl.py - download list of Austrian postal codes
#
# Copyright (C) 2018-2026 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""This download list of postal codes from Austrian Post."""

from __future__ import print_function, unicode_literals

import requests


# The URL of postal codes on the Austrian open-data portal in CSV format.
download_url = 'https://data.rtr.at/api/v1/tables/plz.json'


# The user agent that will be passed in requests
user_agent = 'Mozilla/5.0 (compatible; python-stdnum updater; +https://arthurdejong.org/python-stdnum/)'


# The list of regions that can be used in the document.
regions = {
    'B': 'Burgenland',
    'K': 'Kärnten',
    'N': 'Niederösterreich',
    'O': 'Oberösterreich',
    'Sa': 'Salzburg',
    'St': 'Steiermark',
    'T': 'Tirol',
    'V': 'Vorarlberg',
    'W': 'Wien',
}


if __name__ == '__main__':
    response = requests.get(download_url, timeout=30, headers={'User-Agent': user_agent})
    response.raise_for_status()
    data = response.json()
    # print header
    print('# generated from %s' % download_url)
    print('# version %s published %s' % (
        data['version']['id'], data['version']['published']))
    # build an ordered list of postal codes
    results = set()
    for row in data['data']:
        if row['adressierbar'] == 'Ja':
            results.add((str(row['plz']), row['ort'], regions[row['bundesland']]))
    for code, location, region in sorted(results):
        print('%s location="%s" region="%s"' % (code, location, region))



# tckimlik.py - functions for handling T.C. Kimlik No.
# coding: utf-8
#
# Copyright (C) 2016-2024 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""T.C. Kimlik No. (Turkish personal identification number)

The Turkish Identification Number (Türkiye Cumhuriyeti Kimlik Numarası) is a
unique personal identification number assigned to every citizen of Turkey.
The number consists of 11 digits and the last two digits are check digits.

More information:

* https://en.wikipedia.org/wiki/Turkish_Identification_Number
* https://tckimlik.nvi.gov.tr/

>>> validate('17291716060')
'17291716060'
>>> validate('17291716050')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('1729171606')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('07291716092')  # number must not start with a 0
Traceback (most recent call last):
    ...
InvalidFormat: ...
"""

from __future__ import annotations

from stdnum.exceptions import *
from stdnum.util import clean, get_soap_client, isdigits


tckimlik_wsdl = 'https://tckimlik.nvi.gov.tr/Service/KPSPublic.asmx?WSDL'
"""The WSDL URL of the T.C. Kimlik validation service."""


def compact(number: str) -> str:
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number).strip()


def calc_check_digits(number: str) -> str:
    """Calculate the check digits for the specified number. The number
    passed should not have the check digit included."""
    check1 = (10 - sum((3, 1)[i % 2] * int(n)
              for i, n in enumerate(number[:9]))) % 10
    check2 = (check1 + sum(int(n) for n in number[:9])) % 10
    return '%d%d' % (check1, check2)


def validate(number: str) -> str:
    """Check if the number is a valid T.C. Kimlik number. This checks the
    length and check digits."""
    number = compact(number)
    if not isdigits(number) or number[0] == '0':
        raise InvalidFormat()
    if len(number) != 11:
        raise InvalidLength()
    if calc_check_digits(number) != number[-2:]:
        raise InvalidChecksum()
    return number


def is_valid(number: str) -> bool:
    """Check if the number is a valid T.C. Kimlik number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def check_kps(
    number: str,
    name: str,
    surname: str,
    birth_year: int,
    timeout: float = 30,
    verify: bool | str = True,
) -> bool:  # pragma: no cover
    """Use the T.C. Kimlik validation service to check the provided number.

    Query the online T.C. Kimlik validation service run by the Directorate
    of Population and Citizenship Affairs. The timeout is in seconds. This
    returns a boolean but may raise a SOAP exception for missing or invalid
    values.

    The `timeout` argument specifies the network timeout in seconds.

    The `verify` argument is either a boolean that determines whether the
    server's certificate is validate or a string which must be a path the CA
    certificate bundle to use for verification.
    """
    # this function isn't automatically tested because it would require
    # network access for the tests and unnecessarily load the online service
    number = compact(number)
    client = get_soap_client(tckimlik_wsdl, timeout=timeout, verify=verify)
    result = client.TCKimlikNoDogrula(
        TCKimlikNo=number, Ad=name, Soyad=surname, DogumYili=birth_year)
    if hasattr(result, 'get'):
        return result.get('TCKimlikNoDogrulaResult')  # type: ignore[no-any-return]
    return result  # type: ignore[no-any-return]



# in_.py - functions for handling Japanese Individual Numbers (IN)
# coding: utf-8
#
# Copyright (C) 2025 Luca Sicurello
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""IN (個人番号, kojin bangō, Japanese Individual Number).

The Japanese Individual Number (個人番号, kojin bangō), often referred to as My
number (マイナンバー, mai nambā), is assigned to identify citizens and residents.
The number consists of 12 digits where the last digit is a check digit. No
personal information (such as name, gender, date of birth, etc.) is encoded
in the number.

More information:

 * https://en.wikipedia.org/wiki/Individual_Number
 * https://digitalforensic.jp/2016/03/14/column404/

>>> validate('6214 9832 0257')
'621498320257'
>>> validate('6214 9832 0258')
Traceback (most recent call last):
  ...
InvalidChecksum: ...
>>> validate('6214 9832 025X')
Traceback (most recent call last):
  ...
InvalidFormat: ...
>>> format('621498320257')
'6214 9832 0257'
"""

from __future__ import annotations

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number: str) -> str:
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, '- ').strip()


def calc_check_digit(number: str) -> str:
    """Calculate the check digit. The number passed should not have
    the check digit included."""
    weights = (6, 5, 4, 3, 2, 7, 6, 5, 4, 3, 2)
    s = sum(w * int(n) for w, n in zip(weights, number))
    return str(-s % 11 % 10)


def validate(number: str) -> str:
    """Check if the number is valid. This checks the length and check
    digit."""
    number = compact(number)
    if len(number) != 12:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    if calc_check_digit(number[:-1]) != number[-1]:
        raise InvalidChecksum()
    return number


def is_valid(number: str) -> bool:
    """Check if the number is a valid IN."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number: str) -> str:
    """Reformat the number to the presentation format found on
    My Number Cards."""
    number = compact(number)
    return ' '.join(
        (number[0:4], number[4:8], number[8:12]))



# nik.py - functions for handling Indonesian NIK numbers
# coding: utf-8
#
# Copyright (C) 2024 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""NIK (Nomor Induk Kependudukan, Indonesian identity number).


The Nomor Induk Kependudukan (NIK, Population Identification Number,
sometimes known as Nomor Kartu Tanda Penduduk or Nomor KTP) is issued to
Indonesian citizens.

The number consists of 16 digits in the format PPRRSSDDMMYYXXXX where PPRRSS
(province, city/district, sub-district) indicates the place of residence when
the number was issued. It is followed by a DDMMYY date of birth (for female
40 is added to the day). The last 4 digits are used to make the number
unique.

More information:

* https://id.wikipedia.org/wiki/Nomor_Induk_Kependudukan

>>> validate('3171011708450001')
'3171011708450001'
>>> validate('31710117084500')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('9971011708450001')  # invalid province
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> get_birth_date('3171015708450001')
datetime.date(1945, 8, 17)
>>> get_birth_date('3171012902001234')  # 1900-02-29 doesn't exist
datetime.date(2000, 2, 29)
>>> get_birth_date('3171013002001234')  # 1900-20-30 doesn't exist
Traceback (most recent call last):
    ...
InvalidComponent: ...
"""

from __future__ import annotations

import datetime

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number: str) -> str:
    """Convert the number to the minimal representation.

    This strips the number of any valid separators and removes
    surrounding whitespace.
    """
    return clean(number, ' -.').strip()


def get_birth_date(number: str, minyear: int = 1920) -> datetime.date:
    """Get the birth date from the person's NIK.

    Note that the number only encodes the last two digits of the year so
    this may be a century off.
    """
    number = compact(number)
    day = int(number[6:8]) % 40
    month = int(number[8:10])
    year = int(number[10:12])
    try:
        return datetime.date(year + 1900, month, day)
    except ValueError:
        pass
    try:
        return datetime.date(year + 2000, month, day)
    except ValueError:
        raise InvalidComponent()


def _check_registration_place(number: str) -> dict[str, str]:
    """Use the number to look up the place of registration of the person."""
    from stdnum import numdb
    results = numdb.get('id/loc').info(number[:4])[0][1]
    if not results:
        raise InvalidComponent()
    return results


def validate(number: str) -> str:
    """Check if the number is a valid Indonesian NIK."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) != 16:
        raise InvalidLength()
    get_birth_date(number)
    _check_registration_place(number)
    return number


def is_valid(number: str) -> bool:
    """Check if the number is a valid Indonesian NIK."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# rcs.py - functions for handling French NIR numbers
# coding: utf-8
#
# Copyright (C) 2025 Fabien Michel
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""RCS (French trade registration number for commercial companies).

The RCS number (Registre du commerce et des sociétés) is given by INSEE to a
company with commercial activity when created. It is required for most of
administrative procedures.

The number consists of "RCS" letters followed by name of the city where the company was registered
followed by letter A for a retailer or B for society
followed by the SIREN number

More information:

* https://entreprendre.service-public.fr/vosdroits/F31190
* https://fr.wikipedia.org/wiki/Registre_du_commerce_et_des_sociétés_(France)

>>> validate('RCS Nancy B 323 159 715')
'RCS Nancy B 323159715'
>>> validate('RCS Nancy B 323 159 716')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('RCSNancy B 323 159 716')
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> format('RCS Nancy B323159715')
'RCS Nancy B 323 159 715'
>>> to_siren('RCS Nancy B 323159 715')
'323159715'
"""

from __future__ import annotations

import re

from stdnum.exceptions import *
from stdnum.fr import siren
from stdnum.util import clean


_rcs_validation_regex = re.compile(
    r'^ *(?P<tag>RCS|rcs) +(?P<city>.*?) +(?P<letter>[AB]) *(?P<siren>(?:\d *){9})\b *$',
)


def compact(number: str) -> str:
    """Convert the number to the minimal representation."""
    parts = clean(number).strip().split()
    rest = ''.join(parts[2:])
    return ' '.join(parts[:2] + [rest[:1], siren.compact(rest[1:])])


def validate(number: str) -> str:
    """Validate number is a valid French RCS number."""
    number = compact(number)
    match = _rcs_validation_regex.match(number)
    if not match:
        raise InvalidFormat()
    siren_number = siren.validate(match.group('siren'))
    siren_number = siren.format(siren_number)
    return number


def is_valid(number: str) -> bool:
    """Check if the number provided is valid."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number: str) -> str:
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    match = _rcs_validation_regex.match(number)
    if not match:
        raise InvalidFormat()
    return ' '.join(
        (
            'RCS',
            match.group('city'),
            match.group('letter'),
            siren.format(match.group('siren')),
        ),
    )


def to_siren(number: str) -> str:
    """Extract SIREN number from the RCS number.

    The SIREN number is the 9 last digits of the RCS number.
    """
    number = compact(number)
    match = _rcs_validation_regex.match(clean(number))
    if not match:
        raise InvalidFormat()
    return match.group('siren')



# ogrn.py - functions for handling Russian company registration numbers
# coding: utf-8
#
# Copyright (C) 2024 Ivan Stavropoltsev
# Copyright (C) 2025 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""ОГРН, OGRN, PSRN, ОГРНИП, OGRNIP (Russian Primary State Registration Number).

The ОГРН (Основной государственный регистрационный номер, Primary State
Registration Number) is a Russian identifier for legal entities. The number
consists of 13 or 15 digits and includes information on the type of
organisation, the registration year and a tax inspection code. The 15 digit
variant is called ОГРНИП (Основной государственный регистрационный номер
индивидуального предпринимателя, Primary State Registration Number of an
Individual Entrepreneur).

More information:

* https://ru.wikipedia.org/wiki/Основной_государственный_регистрационный_номер
* https://ru.wikipedia.org/wiki/Основной_государственный_регистрационный_номер_индивидуального_предпринимателя

>>> validate('1022200525819')
'1022200525819'
>>> validate('385768585948949')
'385768585948949'
>>> validate('1022500001328')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

from __future__ import annotations

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number: str) -> str:
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' ')


def calc_check_digit(number: str) -> str:
    """Calculate the control digit of the OGRN based on its length."""
    if len(number) == 13:
        return str(int(number[:-1]) % 11 % 10)
    else:
        return str(int(number[:-1]) % 13)


def validate(number: str) -> str:
    """Determine if the given number is a valid OGRN."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) == 13:
        if number[0] == '0':
            raise InvalidComponent()
    elif len(number) == 15:
        if number[0] not in '34':
            raise InvalidComponent()
    else:
        raise InvalidLength()
    if number[-1] != calc_check_digit(number):
        raise InvalidChecksum()
    return number


def is_valid(number: str) -> bool:
    """Check if the number is a valid OGRN."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# ninea.py - functions for handling Senegal NINEA numbers
# coding: utf-8
#
# Copyright (C) 2023 Leandro Regueiro
# Copyright (C) 2026 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""NINEA (Numéro d'Identification Nationale des Entreprises et Associations, Senegal tax number).

The National Identification Number for Businesses and Associations (NINEA) is
a unique tax identifier for tax purposes in Senegal.

This number consists of 7 digits and is usually followed by a 3 digit tax
identification code called COFI (Code d’Identification Fiscale) that is used
to indicate the company's tax status and legal structure.

More information:

* https://www.wikiprocedure.com/index.php/Senegal%5F-_Obtain_a_Tax_Identification_Number
* https://nkac-audit.com/comprendre-le-ninea-votre-guide-de-lecture-simplifie/
* https://www.creationdentreprise.sn/rechercher-une-societe
* https://www.dci-sn.sn/index.php/obtenir-son-ninea
* https://e-ninea.ansd.sn/search_annuaire

>>> validate('306 7221')
'3067221'
>>> validate('30672212G2')
'30672212G2'
>>> validate('3067221 2G2')
'30672212G2'
>>> validate('3067222')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('1234567 0AZ')
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> format('30672212G2')
'3067221 2G2'
"""  # noqa: E501

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number: str) -> str:
    """Convert the number to the minimal representation.

    This strips the number of any valid separators and removes surrounding
    whitespace.
    """
    return clean(number, ' -/,').upper().strip()


def _validate_cofi(number: str) -> None:
    # The first digit of the COFI indicates the tax status
    # 0: taxpayer subject to the real scheme, not subject to VAT.
    # 1: taxpayer subject to the single global contribution (TOU).
    # 2: taxpayer subject to the real scheme and subject to VAT.
    if number[0] not in '012':
        raise InvalidComponent()
    # The second character is a letter that indicates the tax centre:
    # A: Dakar Plateau 1
    # B: Dakar Plateau 2
    # C: Grand Dakar
    # D: Pikine
    # E: Rufisque
    # F: Thiès
    # G: Centre des grandes Entreprises
    # H: Louga
    # J: Diourbel
    # K: Saint-Louis
    # L: Tambacounda
    # M: Kaolack
    # N: Fatick
    # P: Ziguinchor
    # Q: Kolda
    # R: Prarcelles assainies
    # S: Professions libérales
    # T: Guédiawaye
    # U: Dakar-Medina
    # V: Dakar liberté
    # W: Matam
    # Z: Centre des Moyennes Entreprises
    if number[1] not in 'ABCDEFGHJKLMNPQRSTUVWZ':
        raise InvalidComponent()
    # The third character is a digit that indicates the legal form:
    # 1: Individual-Natural person
    # 2: SARL
    # 3: SA
    # 4: Simple Limited Partnership
    # 5: Share Sponsorship Company
    # 6: GIE
    # 7: Civil Society
    # 8: Partnership
    # 9: Cooperative Association
    # 0: Other
    if number[2] not in '0123456789':
        raise InvalidComponent()


def _checksum(number: str) -> int:
    number = number.zfill(9)
    weights = (1, 2, 1, 2, 1, 2, 1, 2, 1)
    return sum(int(n) * w for n, w in zip(number, weights)) % 10


def validate(number: str) -> str:
    """Check if the number is a valid Senegal NINEA.

    This checks the length and formatting.
    """
    cofi = ''
    number = compact(number)
    if len(number) > 9:
        cofi = number[-3:]
        number = number[:-3]
    if len(number) not in (7, 9):
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    if cofi:
        _validate_cofi(cofi)
    if _checksum(number) != 0:
        raise InvalidChecksum()
    return number + cofi


def is_valid(number: str) -> bool:
    """Check if the number is a valid Senegal NINEA."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number: str) -> str:
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    if len(number) in (7, 9):
        return number
    return ' '.join([number[:-3], number[-3:]])



#!/usr/bin/env python3
# coding: utf-8

# update/cz_banks.py - script to download Bank list from Czech National Bank
#
# Copyright (C) 2022 Petr Přikryl
# Copyright (C) 2026 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""This script downloads the list of banks with bank codes as used in the
IBAN and BIC codes as published by the Czech National Bank."""

import csv
import os.path
from io import StringIO

import requests


# The location of the CSV version of the bank identification codes. Also see
# https://www.cnb.cz/cs/platebni-styk/ucty-kody-bank/
download_url = 'https://www.cnb.cz/cs/platebni-styk/.galleries/ucty_kody_bank/download/kody_bank_CR.csv'


# The user agent that will be passed in requests
user_agent = 'Mozilla/5.0 (compatible; python-stdnum updater; +https://arthurdejong.org/python-stdnum/)'


def get_values(csv_reader):
    """Return values (bank_number, bic, bank_name, certis) from the CSV."""
    # skip first row (header)
    try:
        next(csv_reader)
    except StopIteration:
        pass  # ignore empty CSV

    for row in csv_reader:
        yield row[0], row[2], row[1], row[3] == 'A'


if __name__ == '__main__':
    response = requests.get(download_url, timeout=30, headers={'User-Agent': user_agent})
    response.raise_for_status()
    csv_reader = csv.reader(StringIO(response.content.decode('utf-8')), delimiter=';')
    print('# generated from %s downloaded from' % os.path.basename(download_url))
    print('# %s' % download_url)
    for bank_number, bic, bank, certis in get_values(csv_reader):
        info = '%s' % bank_number
        if bic:
            info += ' bic="%s"' % bic
        if bank:
            info += ' bank="%s"' % bank
        if certis:
            info += ' certis="%s"' % certis
        print(info)



#!/usr/bin/env python3

# update/gs1_ai.py - script to get GS1 application identifiers
#
# Copyright (C) 2019-2026 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""This script downloads GS1 application identifiers from the GS1 web site."""

import datetime
import json
import re

import lxml.html
import requests


# the location of the GS1 application identifiers
download_url = 'https://ref.gs1.org/ai/'


# The user agent that will be passed in requests
user_agent = 'Mozilla/5.0 (compatible; python-stdnum updater; +https://arthurdejong.org/python-stdnum/)'


def fetch_ais():
    """Download application identifiers frm the GS1 website."""
    response = requests.get(download_url, timeout=30, headers={'User-Agent': user_agent})
    document = lxml.html.document_fromstring(response.content)
    element = document.findall('.//script[@type="application/ld+json"]')[0]
    data = json.loads(element.text)
    for entry in data['applicationIdentifiers']:
        if 'applicationIdentifier' in entry:
            ai = entry['applicationIdentifier'].strip()
            formatstring = entry['formatString'].strip()
            if formatstring.startswith(f'N{len(ai)}+'):
                formatstring = formatstring[3:]
            yield (
                ai,
                formatstring,
                entry['separatorRequired'],
                entry['title'].strip(),
                entry['description'].strip())


def group_ai_ranges(ranges):
    """Combine downloaded application identifiers into ranges."""
    first = None
    prev = (None, ) * 5
    for value in sorted(ranges):
        if value[1:] != prev[1:]:
            if first:
                yield (first, *prev)
            first = value[0]
        prev = value
    yield (first, *prev)


if __name__ == '__main__':
    print('# generated from %s' % download_url)
    print('# on %s' % datetime.datetime.now(datetime.UTC))
    for ai1, ai2, format, require_fnc1, name, description in group_ai_ranges(fetch_ais()):
        _type = 'str'
        if re.match(r'^(N[68]\[?\+)?N[0-9]*[.]*[0-9]+\]?$', format) and 'date' in description.lower():
            _type = 'date'
        elif re.match(r'^N[.]*[0-9]+$', format) and 'count' in description.lower():
            _type = 'int'
        ai = ai1
        if ai1 != ai2:
            if len(ai1) == 4:
                _type = 'decimal'
            ai = '%s-%s' % (ai1, ai2)
        print('%s format="%s" type="%s"%s name="%s" description="%s"' % (
            ai, format, _type,
            ' fnc1="1"' if require_fnc1 else '',
            name, description))



# coding: utf-8
# ssn.py - function for handling Belgian social security numbers
#
# Copyright (C) 2023-2024 Jeff Horemans
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""SSN, INSZ, NISS (Belgian social security number).

The Belgian social security identification number, also known as the INSZ
number (Identificatienummer van de sociale zekerheid), NISS number (Numéro
d'identification de la sécurité sociale) or ENSS number (Erkennungsnummer
der sozialen Sicherheit), is the unique identification number of a person
working in or covered for social benefits in Belgium.

For Belgian residents, the number is identical to their Belgian National
Number (Rijksregisternummer, Numéro National).
For non-residents, the number is identical to their Belgian BIS number.

Both numbers consists of 11 digits and encode a person's date of birth and
gender. It encodes the date of birth in the first 6 digits in the format
YYMMDD. The following 3 digits represent a counter of people born on the same
date, separated by sex (odd for male and even for females respectively). The
final 2 digits form a check number based on the 9 preceding digits.


More information:

* https://www.socialsecurity.be/site_nl/employer/applics/belgianidpro/index.htm
* https://overheid.vlaanderen.be/personeel/regelgeving/insz-nummer
* https://www.ksz-bcss.fgov.be/nl/project/rijksregisterksz-registers

>>> compact('85.07.30-033 28')
'85073003328'
>>> compact('98.47.28-997.65')
'98472899765'
>>> validate('85 07 30 033 28')
'85073003328'
>>> validate('98 47 28 997 65')
'98472899765'
>>> validate('17 07 30 033 84')
'17073003384'
>>> validate('01 49 07 001 85')
'01490700185'
>>> validate('12345678901')
Traceback (most recent call last):
    ...
InvalidChecksum: ...

>>> guess_type('85.07.30-033 28')
'nn'
>>> guess_type('98.47.28-997.65')
'bis'

>>> format('85073003328')
'85.07.30-033.28'
>>> get_birth_date('85.07.30-033 28')
datetime.date(1985, 7, 30)
>>> get_birth_year('85.07.30-033 28')
1985
>>> get_birth_month('85.07.30-033 28')
7
>>> get_gender('85.07.30-033 28')
'M'

>>> format('98472899765')
'98.47.28-997.65'
>>> get_birth_date('98.47.28-997.65')
datetime.date(1998, 7, 28)
>>> get_birth_year('98.47.28-997.65')
1998
>>> get_birth_month('98.47.28-997.65')
7
>>> get_gender('98.47.28-997.65')
'M'

"""

from __future__ import annotations

import datetime

from stdnum.be import bis, nn
from stdnum.exceptions import *


_ssn_modules = (nn, bis)


def compact(number: str) -> str:
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return nn.compact(number)


def validate(number: str) -> str:
    """Check if the number is a valid Belgian SSN. This searches for
    the proper sub-type and validates using that."""
    try:
        return bis.validate(number)
    except InvalidComponent:
        # Only try NN validation in case of an invalid component,
        # other validation errors are shared between BIS and NN
        return nn.validate(number)


def is_valid(number: str) -> bool:
    """Check if the number is a valid Belgian SSN number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def guess_type(number: str) -> str | None:
    """Return the Belgian SSN type for which this number is valid."""
    for mod in _ssn_modules:
        if mod.is_valid(number):
            return mod.__name__.rsplit('.', 1)[-1]
    return None


def format(number: str) -> str:
    """Reformat the number to the standard presentation format."""
    return nn.format(number)


def get_birth_year(number: str) -> int | None:
    """Return the year of the birth date."""
    return nn.get_birth_year(number)


def get_birth_month(number: str) -> int | None:
    """Return the month of the birth date."""
    return nn.get_birth_month(number)


def get_birth_date(number: str) -> datetime.date | None:
    """Return the date of birth."""
    return nn.get_birth_date(number)


def get_gender(number: str) -> str | None:
    """Get the person's gender ('M' or 'F')."""
    for mod in _ssn_modules:
        if mod.is_valid(number):
            return mod.get_gender(number)  # type: ignore[no-any-return]
    return None



# bc_phn.py - functions for handling British Columbia Personal Health Numbers (PHNs)
# coding: utf-8
#
# Copyright (C) 2023 Ömer Boratav
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""BC PHN (British Columbia Personal Health Number).

A unique, numerical, lifetime identifier used in the specific identification
of an individual client or patient who has had any interaction with the
British Columbia health system. It is assigned only to and used by one person
and will not be assigned to any other person.

The existence of a PHN does not imply eligibility for health care services in
BC or provide any indication of an individual’s benefit status.

The PNH is a 10-digit number where the first digit is always 9, and the last
digit is a MOD-11 check digit.

More information:

* https://www2.gov.bc.ca/gov/content/health/health-drug-coverage/msp/bc-residents/personal-health-identification
* https://www2.gov.bc.ca/assets/gov/health/practitioner-pro/software-development-guidelines/conformance-standards/vol-4b-app-rules-client-registry.pdf


>>> validate('9698 658 215')
'9698658215'
>>> format('9698658215')
'9698 658 215'
>>> validate('9698648215')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('5736504210')
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> validate('9736A04212')
Traceback (most recent call last):
    ...
InvalidFormat: ...
"""  # noqa: E501

from __future__ import annotations

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number: str) -> str:
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, '- ').strip()


def calc_check_digit(number: str) -> str:
    """Calculate the check digit. The number passed should not have the check
    digit included."""
    weights = (2, 4, 8, 5, 10, 9, 7, 3)
    s = sum((w * int(n)) % 11 for w, n in zip(weights, number))
    return str((11 - s) % 11)


def validate(number: str) -> str:
    """Check if the number is a valid PHN. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if len(number) != 10:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    if number[0] != '9':
        raise InvalidComponent()
    if number[9] != calc_check_digit(number[1:9]):
        raise InvalidChecksum()
    return number


def is_valid(number: str) -> bool:
    """Check if the number is a valid PHN."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number: str) -> str:
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return ' '.join((number[0:4], number[4:7], number[7:]))



# isni.py - functions for handling International Standard Name Identifiers
#
# Copyright (C) 2025 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""ISNI (International Standard Name Identifier).

The International Standard Name Identifier (ISNI) is used to identify
contributors to media content such as books, television programmes, and
newspaper articles. The number consists of 16 digits where the last character
is a check digit.

More information:

* https://en.wikipedia.org/wiki/International_Standard_Name_Identifier
* https://isni.org/

>>> validate('0000 0001 2281 955X')
'000000012281955X'
>>> validate('0000 0001 1111 955X')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('0000 0001 2281')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> format('000000012281955X')
'0000 0001 2281 955X'
"""

from __future__ import annotations

from stdnum.exceptions import *
from stdnum.iso7064 import mod_11_2
from stdnum.util import clean, isdigits


def compact(number: str) -> str:
    """Convert the ISNI to the minimal representation. This strips the number
    of any valid ISNI separators and removes surrounding whitespace."""
    return clean(number, ' -').strip().upper()


def validate(number: str) -> str:
    """Check if the number is a valid ISNI. This checks the length and
    whether the check digit is correct."""
    number = compact(number)
    if not isdigits(number[:-1]):
        raise InvalidFormat()
    if len(number) != 16:
        raise InvalidLength()
    mod_11_2.validate(number)
    return number


def is_valid(number: str) -> bool:
    """Check if the number provided is a valid ISNI."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number: str) -> str:
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return ' '.join((number[:4], number[4:8], number[8:12], number[12:16]))



# bic.py - functions for handling ISO 9362 Business identifier codes
#
# Copyright (C) 2015 Lifealike Ltd
# Copyright (C) 2017-2025 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""BIC (ISO 9362 Business identifier codes).

An ISO 9362 identifier (also: BIC, BEI, or SWIFT code) uniquely
identifies an institution. They are commonly used to route financial
transactions.

The code consists of a 4 letter institution code, a 2 letter country code,
and a 2 character location code, optionally followed by a three character
branch code.

More information:

* https://en.wikipedia.org/wiki/ISO_9362

>>> validate('AGRIFRPP882')
'AGRIFRPP882'
>>> validate('ABNA BE 2A')
'ABNABE2A'
>>> validate('AGRIFRPP')
'AGRIFRPP'
>>> validate('AGRIFRPP8')
Traceback (most recent call last):
    ...
InvalidLength: ..
>>> validate('AGRIF2PP')  # country code can't contain digits
Traceback (most recent call last):
    ...
InvalidFormat: ..
>>> format('agriFRPP')  # conventionally caps
'AGRIFRPP'

"""

from __future__ import annotations

import re

from stdnum.exceptions import *
from stdnum.util import clean


_bic_re = re.compile(r'^[A-Z]{4}(?P<country_code>[A-Z]{2})[0-9A-Z]{2}([0-9A-Z]{3})?$')


# Valid BIC country codes are ISO 3166-1 alpha-2 with the addition of
# XK for the Republic of Kosovo
# It seems that some of the countries included here don't currently have
# any banks with a BIC code (e.g. Antarctica, Iran, Myanmar)
_country_codes = {
    'AD', 'AE', 'AF', 'AG', 'AI', 'AL', 'AM', 'AO', 'AQ', 'AR', 'AS', 'AT', 'AU',
    'AW', 'AX', 'AZ', 'BA', 'BB', 'BD', 'BE', 'BF', 'BG', 'BH', 'BI', 'BJ', 'BL',
    'BM', 'BN', 'BO', 'BQ', 'BR', 'BS', 'BT', 'BV', 'BW', 'BY', 'BZ', 'CA', 'CC',
    'CD', 'CF', 'CG', 'CH', 'CI', 'CK', 'CL', 'CM', 'CN', 'CO', 'CR', 'CU', 'CV',
    'CW', 'CX', 'CY', 'CZ', 'DE', 'DJ', 'DK', 'DM', 'DO', 'DZ', 'EC', 'EE', 'EG',
    'EH', 'ER', 'ES', 'ET', 'FI', 'FJ', 'FK', 'FM', 'FO', 'FR', 'GA', 'GB', 'GD',
    'GE', 'GF', 'GG', 'GH', 'GI', 'GL', 'GM', 'GN', 'GP', 'GQ', 'GR', 'GS', 'GT',
    'GU', 'GW', 'GY', 'HK', 'HM', 'HN', 'HR', 'HT', 'HU', 'ID', 'IE', 'IL', 'IM',
    'IN', 'IO', 'IQ', 'IR', 'IS', 'IT', 'JE', 'JM', 'JO', 'JP', 'KE', 'KG', 'KH',
    'KI', 'KM', 'KN', 'KP', 'KR', 'KW', 'KY', 'KZ', 'LA', 'LB', 'LC', 'LI', 'LK',
    'LR', 'LS', 'LT', 'LU', 'LV', 'LY', 'MA', 'MC', 'MD', 'ME', 'MF', 'MG', 'MH',
    'MK', 'ML', 'MM', 'MN', 'MO', 'MP', 'MQ', 'MR', 'MS', 'MT', 'MU', 'MV', 'MW',
    'MX', 'MY', 'MZ', 'NA', 'NC', 'NE', 'NF', 'NG', 'NI', 'NL', 'NO', 'NP', 'NR',
    'NU', 'NZ', 'OM', 'PA', 'PE', 'PF', 'PG', 'PH', 'PK', 'PL', 'PM', 'PN', 'PR',
    'PS', 'PT', 'PW', 'PY', 'QA', 'RE', 'RO', 'RS', 'RU', 'RW', 'SA', 'SB', 'SC',
    'SD', 'SE', 'SG', 'SH', 'SI', 'SJ', 'SK', 'SL', 'SM', 'SN', 'SO', 'SR', 'SS',
    'ST', 'SV', 'SX', 'SY', 'SZ', 'TC', 'TD', 'TF', 'TG', 'TH', 'TJ', 'TK', 'TL',
    'TM', 'TN', 'TO', 'TR', 'TT', 'TV', 'TW', 'TZ', 'UA', 'UG', 'UM', 'US', 'UY',
    'UZ', 'VA', 'VC', 'VE', 'VG', 'VI', 'VN', 'VU', 'WF', 'WS', 'XK', 'YE', 'YT',
    'ZA', 'ZM', 'ZW',
}


def compact(number: str) -> str:
    """Convert the number to the minimal representation. This strips the
    number of any surrounding whitespace."""
    return clean(number, ' -').strip().upper()


def validate(number: str) -> str:
    """Check if the number is a valid routing number. This checks the length
    and characters in each position."""
    number = compact(number)
    if len(number) not in (8, 11):
        raise InvalidLength()
    match = _bic_re.search(number)
    if not match:
        raise InvalidFormat()
    if match.group('country_code') not in _country_codes:
        raise InvalidComponent()
    return number


def is_valid(number: str) -> bool:
    """Check if the number provided is a valid BIC."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number: str) -> str:
    """Reformat the number to the standard presentation format."""
    return compact(number)



# vid.py - functions for handling Indian personal virtual identity numbers
#
# Copyright (C) 2024 Atul Deolekar
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""VID (Indian personal virtual identity number).

VID is a temporary, revocable 16-digit random number mapped with the Aadhaar number.
VID is used in lieu of Aadhaar number whenever authentication or e-KYC services
are performed.

VID is made up of 16 digits where the last digits is a check digit
calculated using the Verhoeff algorithm. The numbers are generated in a
random, non-repeating sequence and do not begin with 0 or 1.

More information:

* https://uidai.gov.in/en/contact-support/have-any-question/284-faqs/aadhaar-online-services/virtual-id-vid.html
* https://uidai.gov.in/images/resource/UIDAI_Circular_11012018.pdf

>>> validate('2341234123412341')
'2341234123412341'
>>> validate('2341234123412342')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('1341234123412341')  # number should not start with 0 or 1
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> validate('13412341234123')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('2222222222222222')  # number cannot be a palindrome
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> format('2341234123412342')
'2341 2341 2341 2342'
>>> mask('2341234123412342')
'XXXX XXXX  XXXX 2342'
"""

from __future__ import annotations

import re

from stdnum import verhoeff
from stdnum.exceptions import *
from stdnum.util import clean


_vid_re = re.compile(r'^[2-9][0-9]{15}$')
"""Regular expression used to check syntax of VID numbers."""


def compact(number: str) -> str:
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').strip()


def validate(number: str) -> str:
    """Check if the number provided is a valid VID number. This checks
    the length, formatting and check digit."""
    number = compact(number)
    if len(number) != 16:
        raise InvalidLength()
    if not _vid_re.match(number):
        raise InvalidFormat()
    if number == number[::-1]:
        raise InvalidFormat()  # VID cannot be a palindrome
    verhoeff.validate(number)
    return number


def is_valid(number: str) -> bool:
    """Check if the number provided is a valid VID number. This checks
    the length, formatting and check digit."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number: str) -> str:
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return ' '.join((number[:4], number[4:8], number[8:12], number[12:]))


def mask(number: str) -> str:
    """Masks the first 8 digits as per Ministry of Electronics and
    Information Technology (MeitY) guidelines."""
    number = compact(number)
    return 'XXXX XXXX XXXX ' + number[-4:]



# ecnumber.py - functions for handling European Community Numbers

# Copyright (C) 2023 Daniel Weber
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""EC Number (European Community number).

The EC Number is a unique seven-digit number assigned to chemical substances
for regulatory purposes within the European Union by the European Commission.

More information:

* https://en.wikipedia.org/wiki/European_Community_number

>>> validate('200-001-8')
'200-001-8'
>>> validate('200-001-9')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('20-0001-8')
Traceback (most recent call last):
    ...
InvalidFormat: ...
"""

from __future__ import annotations

import re

from stdnum.exceptions import *
from stdnum.util import clean


_ec_number_re = re.compile(r'^[0-9]{3}-[0-9]{3}-[0-9]$')


def compact(number: str) -> str:
    """Convert the number to the minimal representation."""
    number = clean(number, ' ').strip()
    if '-' not in number:
        number = '-'.join((number[:3], number[3:6], number[6:]))
    return number


def calc_check_digit(number: str) -> str:
    """Calculate the check digit for the number. The passed number should not
    have the check digit included."""
    number = compact(number).replace('-', '')
    return str(
        sum((i + 1) * int(n) for i, n in enumerate(number)) % 11)[0]


def validate(number: str) -> str:
    """Check if the number provided is a valid EC Number."""
    number = compact(number)
    if not len(number) == 9:
        raise InvalidLength()
    if not _ec_number_re.match(number):
        raise InvalidFormat()
    if number[-1] != calc_check_digit(number[:-1]):
        raise InvalidChecksum()
    return number


def is_valid(number: str) -> bool:
    """Check if the number provided is a valid EC Number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# voen.py - functions for handling Azerbaijan VOEN numbers
# coding: utf-8
#
# Copyright (C) 2022 Leandro Regueiro
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""VÖEN (Vergi ödəyicisinin eyniləşdirmə nömrəsi, Azerbaijan tax number).

The Vergi ödəyicisinin eyniləşdirmə nömrəsi is issued by the Azerbaijan state
tax authorities to individuals and legal entities.

This number consists of 10 digits. The first two digits are the code for the
territorial administrative unit. The following six digits are a serial
number. The ninth digit is a check digit. The tenth digit represents the
legal status of a taxpayer: 1 for legal persons and 2 for natural persons.

More information:

* https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/Azerbaijan-TIN.pdf
* https://www.e-taxes.gov.az/ebyn/payerOrVoenChecker.jsp

>>> validate('140 155 5071')
'1401555071'
>>> validate('140 155 5081')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('1400057424')
Traceback (most recent call last):
    ...
InvalidComponent: ...
"""  # noqa: E501

from __future__ import annotations

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number: str) -> str:
    """Convert the number to the minimal representation.

    This strips the number of any valid separators and removes surrounding
    whitespace.
    """
    number = clean(number, ' ')
    if len(number) == 9:
        number = '0' + number
    return number


def _calc_check_digit(number: str) -> str:
    """Calculate the check digit for the VÖEN."""
    weights = [4, 1, 8, 6, 2, 7, 5, 3]
    return str(sum(w * int(n) for w, n in zip(weights, number)) % 11)


def validate(number: str) -> str:
    """Check if the number is a valid Azerbaijan VÖEN number."""
    number = compact(number)
    if len(number) != 10:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    if number[-1] not in ('1', '2'):
        raise InvalidComponent()
    if number[-2:-1] != _calc_check_digit(number):
        raise InvalidChecksum()
    return number


def is_valid(number: str) -> bool:
    """Check if the number is a valid Azerbaijan VÖEN number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# test_eu_vat.py - functions for testing the online VIES VAT validation
# coding: utf-8
#
# Copyright (C) 2017 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

# This is a separate test file because it should not be run regularly
# because it could negatively impact the VIES service.

"""Extra tests for the stdnum.eu.vat module."""

import os
import unittest

from stdnum.eu import vat


@unittest.skipIf(
    not os.environ.get('ONLINE_TESTS'),
    'Do not overload online services')
class TestVies(unittest.TestCase):
    """Test the VIES web service provided by the European commission for
    validation VAT numbers of European countries."""

    def test_check_vies(self) -> None:
        """Test stdnum.eu.vat.check_vies()"""
        result = vat.check_vies('NL4495445B01')
        self.assertTrue(result['valid'])
        self.assertEqual(result['countryCode'], 'NL')
        self.assertEqual(result['vatNumber'], '004495445B01')

    def test_check_vies_approx(self) -> None:
        """Test stdnum.eu.vat.check_vies_approx()"""
        result = vat.check_vies_approx('NL4495445B01', 'NL4495445B01')
        self.assertTrue(result['valid'])
        self.assertEqual(result['countryCode'], 'NL')
        self.assertEqual(result['vatNumber'], '004495445B01')



# leitweg.py - functions for handling Leitweg-ID
# coding: utf-8
#
# Copyright (C) 2025 Holvi Payment Services Oy
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Leitweg-ID, a buyer reference or routing identifier for electronic invoices.

For the successful transmission of an electronic invoice as invoicing party or
sender, a unique identification and addressing of the invoice recipient is
required. The Leitweg-ID must be transmitted as a mandatory requirement for
electronic invoicing to public contracting authorities in the federal
administration.

More information:

* https://leitweg-id.de/
* https://de.wikipedia.org/wiki/Leitweg-ID
* https://xeinkauf.de/app/uploads/2022/11/Leitweg-ID-Formatspezifikation-v2-0-2-1.pdf

>>> validate('991-03730-19')
'991-03730-19'
>>> validate('1-03730-19')
Traceback (most recent call last):
    ...
InvalidFormat: ...
"""

from __future__ import annotations

import re

from stdnum.exceptions import *
from stdnum.iso7064 import mod_97_10


_pattern = re.compile(r'[0-9]{2,12}(-[0-9A-Z]{,30})?-[0-9]{2}')


def compact(number: str) -> str:
    """
    Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace.
    """
    return number.strip().upper()  # no valid separators, dashes part of the format


def validate(number: str) -> str:
    """
    Check if the number provided is valid. This checks the format, state or
    federal government code, and check digits.
    """
    if not isinstance(number, str):
        raise InvalidFormat()
    number = compact(number)
    # 2.1 Bestandteile der Leitweg-ID
    if not 5 <= len(number) <= 46:
        raise InvalidLength()
    # 2.1 Bestandteile der Leitweg-ID
    if not re.fullmatch(_pattern, number):
        raise InvalidFormat()
    # 2.2.1 Kennzahl des Bundeslandes/des Bundes
    if not number[:2] in {
        '01', '02', '03', '04', '05', '06', '07', '08', '09', '10',
        '11', '12', '13', '14', '15', '16', '99',
    }:
        raise InvalidComponent()
    # 2.4 Prüfziffer
    mod_97_10.validate(number.replace('-', ''))
    return number


def is_valid(number: str) -> bool:
    """Check if the number provided is valid. This checks the length and
    check digit."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# cvr.py - functions for handling Danish CVR numbers
#
# Copyright (C) 2012-2015 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""CVR (Momsregistreringsnummer, Danish VAT number).

The CVR (Momsregistreringsnummer, VAT) is an 8 digit number with a
straightforward check mechanism.

>>> validate('DK 13585628')
'13585628'
>>> validate('DK 13585627')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

from __future__ import annotations

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number: str) -> str:
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    number = clean(number, ' -.,/:').upper().strip()
    if number.startswith('DK'):
        number = number[2:]
    return number


def checksum(number: str) -> int:
    """Calculate the checksum."""
    weights = (2, 7, 6, 5, 4, 3, 2, 1)
    return sum(w * int(n) for w, n in zip(weights, number)) % 11


def validate(number: str) -> str:
    """Check if the number provided is a valid VAT number. This checks the
    length, formatting and check digit."""
    number = compact(number)
    if not isdigits(number) or number[0] == '0':
        raise InvalidFormat()
    if len(number) != 8:
        raise InvalidLength()
    if checksum(number) != 0:
        raise InvalidChecksum()
    return number


def is_valid(number: str) -> bool:
    """Check if the number provided is a valid VAT number. This checks the
    length, formatting and check digit."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



#!/usr/bin/env python3

# check_headers.py - check that all source files have licensing information
#
# Copyright (C) 2023 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""This script checks that all source files have licensing information."""

import glob
import os
import re
import sys
import textwrap


# Regext to match the first line
identification_re = re.compile(
    r'^(#!/usr/bin/env python3?\s+|/\*\s+)?'
    r'(# coding: utf-8\s+)?'
    r'(\s?#\s+)?'
    r'(?P<filename>[^ :]+) - ', re.MULTILINE)

# Regex to match standard license blurb
license_re = re.compile(textwrap.dedent(r'''
    [# ]*This library is free software; you can redistribute it and/or
    [# ]*modify it under the terms of the GNU Lesser General Public
    [# ]*License as published by the Free Software Foundation; either
    [# ]*version 2.1 of the License, or .at your option. any later version.
    [# ]*
    [# ]*This library is distributed in the hope that it will be useful,
    [# ]*but WITHOUT ANY WARRANTY; without even the implied warranty of
    [# ]*MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    [# ]*Lesser General Public License for more details.
    [# ]*
    [# ]*You should have received a copy of the GNU Lesser General Public
    [# ]*License along with this library; if not, write to the Free Software
    [# ]*Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    [# ]*02110-1301 USA
    ''').strip(), re.MULTILINE)


def get_file_header(filename):
    """Read the file header from the file."""
    with open(filename, 'rt') as f:
        # Only read the first 2048 bytes to avoid loading too much and the
        # license information should be in the first part anyway.
        return f.read(2048)


if __name__ == '__main__':
    files_to_check = (
        glob.glob('*.py') +
        glob.glob('stdnum/**/*.py', recursive=True) +
        glob.glob('tests/**/*.doctest', recursive=True) +
        glob.glob('scripts/*', recursive=True) +
        glob.glob('update/**/*.py', recursive=True) +
        glob.glob('online_check/*.wsgi', recursive=True) +
        glob.glob('online_check/check.js', recursive=True)
    )

    # Look for files with incorrect first lines or license
    fail = False
    for filename in sorted(files_to_check):
        contents = get_file_header(filename)
        m = identification_re.match(contents)
        if not bool(m) or m.group('filename') not in (filename, os.path.basename(filename)):
            print('%s: Incorrect file identification' % filename)
            fail = True
        if not bool(license_re.search(contents)):
            print('%s: Incorrect license text' % filename)
            fail = True

    if fail:
        sys.exit(1)



#!/usr/bin/env python3

# update/cn_loc.py - script to fetch data from the CN Open Data community
#
# Copyright (C) 2014-2015 Jiangge Zhang
# Copyright (C) 2015-2026 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""This downloads the birth place codes from from Wikipedia."""

import re
import unicodedata
from collections import defaultdict

import requests


# The wikipedia pages to download
wikipedia_pages = [f'中华人民共和国行政区划代码 ({i}区)' for i in range(1, 9)]


# The user agent that will be passed in requests
user_agent = 'Mozilla/5.0 (compatible; python-stdnum updater; +https://arthurdejong.org/python-stdnum/)'


def get_wikipedia_url(page):
    """Get the Simplified Chinese Wikipedia page URL."""
    return f'https://zh.wikipedia.org/w/index.php?title={page.replace(" ", "_")}&action=raw'  # noqa: E231


# Regular expression for matching province heading
province_re = re.compile(r'^== *(?P<province>.*) +\((?P<prefix>[0-9]+)\) +==')

# Regular expression for matching table row
entry_re = re.compile(
    r'^\| *(?P<number>[0-9]{6}) *' +
    r'\|\| *(?P<activation>.*) *' +
    r'\|\| *(?P<revocation>.*) *' +
    r'\|\| *(?P<county>.*) *' +
    r'\|\| *(?P<code>.*)')


def clean(value):
    """Normalise (partially) unicode strings."""
    # Remove unicode parenthesis that include space with normal ones
    value = value.replace(
        unicodedata.lookup('FULLWIDTH LEFT PARENTHESIS'), ' (',
    ).replace(
        unicodedata.lookup('FULLWIDTH RIGHT PARENTHESIS'), ') ',
    )
    # Remove Wikipedia links
    return re.sub(r'\[\[([^]|]*\|)?([^]|]+)\]\]', r'\2', value)


def parse_county(county, activation, revocation):
    """Parse the county string and return ranges counties."""
    for value in county.split('<br>'):
        m = re.match(r'(?P<county>.*) +\((?P<year>[0-9]{4})年至今\) +', value)
        # This parses various formats as seen on Wikipedia
        if m:  # starting with year
            yield f'[{m.group("year")}-{revocation}]{m.group("county")}'
            continue
        m = re.match(r'(?P<county>.*) +\((?P<year>[0-9]{4})年前\) +', value)
        if m:  # before given year
            yield f'[{activation}-{int(m.group("year")) - 1}]{m.group("county")}'
            continue
        m = re.match(r'(?P<county>.*) +\((?P<years>[0-9]{4}-[0-9]{4})年曾撤销\) +', value)
        if m:  # abolished between years
            if activation or revocation:
                yield f'[{activation}-{revocation}]{m.group("county")}'
            else:
                yield m.group('county')
            continue
        m = re.match(r'(?P<county>.*) +\((?P<start>[0-9]{4})年?-(?P<end>[0-9]{4})年\) +', value)
        if m:
            yield f'[{m.group("start")}-{int(m.group("end")) - 1}]{m.group("county")}'
            continue
        if activation or revocation:
            yield f'[{activation}-{revocation}]{value}'
        else:
            yield value


def parse_page(content):
    """Parse the contents of the Wikipedia page and return number, county, province tuples."""
    province = None
    prefix = None
    for line in clean(content).splitlines():
        line = clean(line)
        m = province_re.match(line)
        if m:
            province = m.group('province')
            prefix = m.group('prefix')
            continue
        m = entry_re.match(line)
        if m:
            number = m.group('number')
            assert number.startswith(prefix)
            counties = m.group('county')
            try:
                activation = str(int(m.group('activation')))
            except ValueError:
                activation = ''
            try:
                revocation = str(int(m.group('revocation')))
            except ValueError:
                revocation = ''
            for county in parse_county(counties, activation, revocation):
                yield prefix, province, number, county.strip()


if __name__ == '__main__':
    """Output a data file in the right format."""
    print('# Downloaded from')
    for page in wikipedia_pages:
        print(f'# {get_wikipedia_url(page)}')
    # Download all data
    provinces = {}
    numbers = defaultdict(lambda: defaultdict(list))
    for page in wikipedia_pages:
        response = requests.get(get_wikipedia_url(page), timeout=30, headers={'User-Agent': user_agent})
        response.raise_for_status()
        for prefix, province, number, county in parse_page(response.text):
            provinces[prefix] = province
            numbers[prefix][number].append(county)
    # Print data
    for prefix, province in sorted(provinces.items()):
        print(f'{prefix} province="{province}"')
        for number, counties in sorted(numbers[prefix].items()):
            county = ','.join(sorted(counties))
            print(f'  {number[2:]} county="{county}"')



# identiteitskaartnummer.py - functions for handling Dutch passport numbers
# coding: utf-8
#
# Copyright (C) 2024 Jeff Horemans
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Identiteitskaartnummer, Paspoortnummer (the Dutch passport number).

Each Dutch passport has an unique number of 9 alphanumerical characters.
The first 2 characters are always letters and the last character is a number.
The 6 characters in between can be either.

The letter "O" is never used to prevent it from being confused with the number "0".
Zeros are allowed, but are not used in numbers issued after December 2019.

More information:

* https://www.rvig.nl/node/356
* https://nl.wikipedia.org/wiki/Paspoort#Nederlands_paspoort

>>> compact('EM0000000')
'EM0000000'
>>> compact('XR 1001R5 8')
'XR1001R58'
>>> validate('EM0000000')
'EM0000000'
>>> validate('XR1001R58')
'XR1001R58'
>>> validate('XR1001R')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('581001RXR')
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> validate('XR1O01R58')
Traceback (most recent call last):
    ...
InvalidComponent: ...
"""

from __future__ import annotations

import re

from stdnum.exceptions import *
from stdnum.util import clean


def compact(number: str) -> str:
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding white space."""
    return clean(number, ' ').strip().upper()


def validate(number: str) -> str:
    """Check if the number is a valid passport number.
    This checks the length, formatting and check digit."""
    number = compact(number)
    if len(number) != 9:
        raise InvalidLength()
    if not re.match('^[A-Z]{2}[0-9A-Z]{6}[0-9]$', number):
        raise InvalidFormat()
    if 'O' in number:
        raise InvalidComponent(InvalidComponent("The letter 'O' is not allowed"))
    return number


def is_valid(number: str) -> bool:
    """Check if the number is a valid Dutch passport number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# excise.py - functions for handling EU Excise numbers
# coding: utf-8
#
# Copyright (C) 2023 Cédric Krier
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""European Excise Number

The excise duty identification number assigned to businesses authorised to
operate with excise goods (e.g. alcohol, tobacco, energy products, etc.).

The number is issued by the national customs or tax authority of the Member
State where the business is established. The number consists of a two-letter
country code followed by up to 13 alphanumeric characters.

More information:

* https://ec.europa.eu/taxation_customs/dds2/seed/

>>> validate('LU 987ABC')
'LU00000987ABC'
"""

from __future__ import annotations

from stdnum.eu.vat import MEMBER_STATES
from stdnum.exceptions import *
from stdnum.util import NumberValidationModule, clean, get_cc_module


_country_modules = dict()


def _get_cc_module(cc: str) -> NumberValidationModule | None:
    """Get the Excise number module based on the country code."""
    cc = cc.lower()
    if cc not in MEMBER_STATES:
        raise InvalidComponent()
    if cc not in _country_modules:
        _country_modules[cc] = get_cc_module(cc, 'excise')
    return _country_modules[cc]


def compact(number: str) -> str:
    """Convert the number to the minimal representation. This strips the number
    of any valid separators and removes surrounding whitespace."""
    number = clean(number, ' ').upper().strip()
    if len(number) < 13:
        number = number[:2] + number[2:].zfill(11)
    return number


def validate(number: str) -> str:
    """Check if the number is a valid Excise number."""
    number = compact(number)
    if len(number) != 13:
        raise InvalidLength()
    module = _get_cc_module(number[:2])
    if module:
        module.validate(number)
    return number


def is_valid(number: str) -> bool:
    """Check if the number is a valid excise number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False
